/*
 This file is part of the digitalSTROM ds485-core library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-core library is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-core library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-core library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/**
 * @file
 * Board Support Package, needed for qp
 */
#include "bsp.h"

#include "qp_port.h"
#include "digitalSTROM/ds485.h"
#include "ds485-signals.h"
#include "ds485-qp.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>                            /* for memcpy() and memset() */
#include <sys/select.h>
#include <termios.h>
#include <unistd.h>

#ifdef Q_SPY
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <errno.h>

    static int l_sock = -1;
#endif

Q_DEFINE_THIS_FILE

/*..........................................................................*/
#ifdef Q_SPY
static void *idleThread(void *me) {      /* the expected P-Thread signature */
  struct timespec time = { 0 };
    for (;;) {
        time.tv_nsec = 10000e3;
        nanosleep(&time, NULL);
	QS_FLUSH();
    }

    return (void *)0;                                     /* return success */
}
#endif
/*..........................................................................*/
void BSP_init()
{
#ifdef Q_SPY
    char const *hostAndPort = "localhost:6601";

    if (!QS_INIT(hostAndPort)) {
        printf("\nUnable to open QS socket\n");
        QF_stop();
    }
    printf("QP/C DS485 Sample application"
           "\nQEP %s\nQF  %s\n"
           "Press ESC to quit...\n",
           QEP_getVersion(),
           QF_getVersion());
#endif
}
/*..........................................................................*/
void QF_onStartup(void) {                               /* startup callback */
#ifdef Q_SPY
    pthread_attr_t attr;
    struct sched_param param;
    pthread_t idle;

    /* SCHED_FIFO corresponds to real-time preemptive priority-based scheduler
    * NOTE: This scheduling policy requires the superuser priviledges
    */
    pthread_attr_init(&attr);
    pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
    param.sched_priority = sched_get_priority_min(SCHED_FIFO);

    pthread_attr_setschedparam(&attr, &param);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    if (pthread_create(&idle, &attr, &idleThread, 0) != 0) {
               /* Creating the p-thread with the SCHED_FIFO policy failed.
               * Most probably this application has no superuser privileges,
               * so we just fall back to the default SCHED_OTHER policy
               * and priority 0.
               */
        pthread_attr_setschedpolicy(&attr, SCHED_OTHER);
        param.sched_priority = 0;
        pthread_attr_setschedparam(&attr, &param);
        Q_ALLEGE(pthread_create(&idle, &attr, &idleThread, 0) == 0);
    }
    pthread_attr_destroy(&attr);
#endif

}
/*..........................................................................*/
void QF_onCleanup(void) {                               /* cleanup callback */
#ifdef Q_SPY
    printf("\nBye! Bye!\n");
#endif
    QS_EXIT();                                    /* perfomr the QS cleanup */
}
/*..........................................................................*/
void Q_onAssert(char const Q_ROM * const Q_ROM_VAR file, int line) {
    fprintf(stderr, "Assertion failed in %s, line %d\n", file, line);
    fprintf(stderr, "Queues:\n");
    int i;
    for (i = 0; i < QF_MAX_ACTIVE; i++) {
        QActive *active = QF_active_[i];
        if (!active) {
            continue;
        }
        int j;
        for (j = 0; j < active->queuesSize; j++) {
            QEQueue *queue = active->queues[j];
            fprintf(stderr, "%12s/%20s min:%3d free:%3d size:%3d\n",
                active->fileName, queue->name,
                queue->nMin, queue->nFree, queue->end);
        }
    }
    QF_stop();
    abort();
}

/*--------------------------------------------------------------------------*/
#ifdef Q_SPY                                         /* define QS callbacks */

/*..........................................................................*/
uint8_t QS_onStartup(void const *arg) {
    static uint8_t qsBuf[1024];                /* 1K buffer for Quantum Spy */
    char host[64];
    char const *src;
    char *addr;
    uint16_t port = 6601;                                   /* default port */
    struct sockaddr_in servAddr;
    struct hostent *server;

    QS_initBuf(qsBuf, sizeof(qsBuf));

    src = (char const *)arg;
    addr = host;
    while ((*src != '\0') && (*src != ':') && (addr < &host[sizeof(host)])) {
        *addr++ = *src++;
    }
    *addr = '\0';
    if (*src == ':') {
        port = (uint16_t)strtoul(src + 1, NULL, 10);
    }

    l_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);       /* TCP socket */
    if (l_sock == -1){
        printf("Socket cannot be created.\n"
               "socket error 0x%08X.",
               errno);
        return (uint8_t)0;
    }

    server = gethostbyname(host);
    if (server == NULL) {
        printf("QSpy host name cannot be resolved.\n"
               "socket error 0x%08X.",
               errno);
        return (uint8_t)0;
    }
    memset(&servAddr, 0, sizeof(servAddr));
    servAddr.sin_family = AF_INET;
    memcpy(&servAddr.sin_addr, server->h_addr, server->h_length);
    servAddr.sin_port = htons(port);
    if (connect(l_sock, (struct sockaddr *)&servAddr, sizeof(servAddr)) == -1)
    {
        printf("Socket cannot be connected to the QSpy server.\n"
               "socket error 0x%08X.",
               errno);
        QS_onCleanup();
        return (uint8_t)0;
    }

    /* setup the QS filters... */
    QS_FILTER_OFF(QS_ALL_RECORDS);
    /* QS_FILTER_ON(QS_ALL_RECORDS); */

    /*
     * see src/include/ds485-signals.h
     * for all ds485 specific events
     */

#if 1
    /* only observer main state machine */
    QS_FILTER_SM_OBJ(AO_DS485);
    QS_FILTER_AO_OBJ(AO_DS485);
    QS_FILTER_MP_OBJ(AO_DS485);
    QS_FILTER_EQ_OBJ(AO_DS485);
    QS_FILTER_TE_OBJ(AO_DS485);
    QS_FILTER_AP_OBJ(AO_DS485);
#endif

    /*
     * see src/qpc/include/qs.h
     * for all generic qep events
     */

#if 1
    /* QEP machine states */
    QS_FILTER_ON(QS_QEP_STATE_EMPTY);
    QS_FILTER_ON(QS_QEP_STATE_ENTRY);
    QS_FILTER_ON(QS_QEP_STATE_EXIT);
    QS_FILTER_ON(QS_QEP_STATE_INIT);
    QS_FILTER_ON(QS_QEP_INIT_TRAN);
    QS_FILTER_OFF(QS_QEP_INTERN_TRAN);
    QS_FILTER_ON(QS_QEP_TRAN);
    QS_FILTER_OFF(QS_QEP_IGNORED);
    QS_FILTER_OFF(QS_QEP_DISPATCH);
#endif

#if 0
    /* QEP framework states */
    QS_FILTER_ON(QS_QF_ACTIVE_ADD);
    QS_FILTER_ON(QS_QF_ACTIVE_REMOVE);
    QS_FILTER_ON(QS_QF_ACTIVE_SUBSCRIBE);
    QS_FILTER_ON(QS_QF_ACTIVE_UNSUBSCRIBE);
    QS_FILTER_ON(QS_QF_ACTIVE_POST_FIFO);
    QS_FILTER_ON(QS_QF_ACTIVE_POST_LIFO);
    QS_FILTER_ON(QS_QF_ACTIVE_GET);
    QS_FILTER_ON(QS_QF_ACTIVE_GET_LAST);
    QS_FILTER_ON(QS_QF_EQUEUE_INIT);
    QS_FILTER_ON(QS_QF_EQUEUE_POST_FIFO);
    QS_FILTER_ON(QS_QF_EQUEUE_POST_LIFO);
    QS_FILTER_ON(QS_QF_EQUEUE_GET);
    QS_FILTER_ON(QS_QF_EQUEUE_GET_LAST);
    QS_FILTER_ON(QS_QF_MPOOL_INIT);
    QS_FILTER_ON(QS_QF_MPOOL_GET);
    QS_FILTER_ON(QS_QF_MPOOL_PUT);
    QS_FILTER_ON(QS_QF_PUBLISH);
    QS_FILTER_ON(QS_QF_NEW);
    QS_FILTER_ON(QS_QF_GC_ATTEMPT);
    QS_FILTER_ON(QS_QF_GC);
    QS_FILTER_ON(QS_QF_TICK);
    QS_FILTER_ON(QS_QF_TIMEEVT_ARM);
    QS_FILTER_ON(QS_QF_TIMEEVT_AUTO_DISARM);
    QS_FILTER_ON(QS_QF_TIMEEVT_DISARM_ATTEMPT);
    QS_FILTER_ON(QS_QF_TIMEEVT_DISARM);
    QS_FILTER_ON(QS_QF_TIMEEVT_REARM);
    QS_FILTER_ON(QS_QF_TIMEEVT_POST);
    QS_FILTER_ON(QS_QF_INT_LOCK);
    QS_FILTER_ON(QS_QF_INT_UNLOCK);
    QS_FILTER_ON(QS_QF_ISR_ENTRY);
    QS_FILTER_ON(QS_QF_ISR_EXIT);

    /* QEP kernel -- we are using pthread wrapper */
    QS_FILTER_ON(QS_QK_MUTEX_LOCK);
    QS_FILTER_ON(QS_QK_MUTEX_UNLOCK);
    QS_FILTER_ON(QS_QK_SCHEDULE);
#endif

    /* QEP spy states */
    QS_FILTER_ON(QS_SIG_DICTIONARY);
    QS_FILTER_ON(QS_OBJ_DICTIONARY);
    QS_FILTER_ON(QS_FUN_DICTIONARY);

    return (uint8_t)1;                                           /* success */
}
/*..........................................................................*/
void QS_onCleanup(void) {
    if (l_sock != -1) {
        close(l_sock);
    }
}
/*..........................................................................*/
QSTimeCtr QS_onGetTime(void) {           /* invoked within critical section */
    return (QSTimeCtr)clock();
}
/*..........................................................................*/
void QS_onFlush(void) {
    uint16_t nBytes = 1000;
    uint8_t const *block;
    QF_INT_LOCK(dummy);
    while ((block = QS_getBlock(&nBytes)) != (uint8_t *)0) {
        send(l_sock, (char const *)block, nBytes, 0);
        nBytes = 1000;
    }
    QF_INT_UNLOCK(dummy);
}
#endif                                                             /* Q_SPY */
/*--------------------------------------------------------------------------*/
